# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for PLDPmf."""

import unittest

from absl.testing import parameterized
import numpy as np

from dp_accounting.pld import common
from dp_accounting.pld import pld_pmf
from dp_accounting.pld import test_util


class PLDPmfTest(parameterized.TestCase):

  def _create_pmf(self,
                  discretization: float,
                  dense: bool,
                  infinity_mass: float = 0.0,
                  lower_loss: int = 0,
                  probs: np.ndarray = np.array([1.0]),
                  pessimistic_estimate: bool = True) -> pld_pmf.PLDPmf:
    """Helper function for creating PLD for testing."""
    if dense:
      return pld_pmf.DensePLDPmf(discretization, lower_loss, probs,
                                 infinity_mass, pessimistic_estimate)

    loss_probs = common.list_to_dictionary(probs, lower_loss)
    return pld_pmf.SparsePLDPmf(loss_probs, discretization, infinity_mass,
                                pessimistic_estimate)

  def _check_dense_probs(self, dense_pmf: pld_pmf.DensePLDPmf,
                         expected_lower_loss: int, expected_probs: np.ndarray):
    """Checks that resulting dense pmf satisfies expectations."""
    self.assertEqual(expected_lower_loss, dense_pmf._lower_loss)
    self.assertSequenceAlmostEqual(expected_probs, dense_pmf._probs)

  def _check_sparse_probs(self, sparse_pmf: pld_pmf.SparsePLDPmf,
                          expected_lower_loss: int, expected_probs: np.ndarray):
    """Checks that resulting sparse pmf satisfies expectations."""
    expected_loss_probs = common.list_to_dictionary(expected_probs,
                                                    expected_lower_loss)
    test_util.assert_dictionary_almost_equal(self, expected_loss_probs,
                                             sparse_pmf._loss_probs)

  def _check_sparse_pld_pmf_equal(self, sparse_pmf: pld_pmf.SparsePLDPmf,
                                  discretization: float, infinity_mass: float,
                                  loss_probs: dict[int, float],
                                  pessimistic_estimate: bool):
    self.assertEqual(discretization, sparse_pmf._discretization)
    self.assertAlmostEqual(infinity_mass, sparse_pmf._infinity_mass)
    self.assertEqual(pessimistic_estimate, sparse_pmf._pessimistic_estimate)
    test_util.assert_dictionary_almost_equal(self, loss_probs,
                                             sparse_pmf._loss_probs)

  @parameterized.parameters(False, True)
  def test_delta_for_epsilon(self, dense: bool):
    discretization = 0.1
    infinity_mass = 0.1
    lower_loss = -1
    probs = np.array([0.2, 0.3, 0, 0.4])
    pmf = self._create_pmf(discretization, dense, infinity_mass, lower_loss,
                           probs)
    self.assertAlmostEqual(0.1, pmf.get_delta_for_epsilon(3))  # infinity_mass
    self.assertAlmostEqual(0.1 + 0.4 * (1 - np.exp(-0.1)),
                           pmf.get_delta_for_epsilon(0.1))
    self.assertAlmostEqual(1, pmf.get_delta_for_epsilon(-20))
    self.assertEqual(infinity_mass, pmf.get_delta_for_epsilon(np.inf))
    self.assertAlmostEqual(1, pmf.get_delta_for_epsilon(-np.inf))

  @parameterized.parameters(False, True)
  def test_delta_for_epsilon_vectorized(self, dense: bool):
    discretization = 0.1
    infinity_mass = 0.1
    lower_loss = -1
    probs = np.array([0.2, 0.3, 0, 0.4])
    pmf = self._create_pmf(discretization, dense, infinity_mass, lower_loss,
                           probs)
    epsilon = [-np.inf, -20, 0.1, np.inf]
    expected_delta = [
        1, 1, infinity_mass + 0.4 * (1 - np.exp(-0.1)), infinity_mass
    ]

    self.assertSequenceAlmostEqual(expected_delta,
                                   pmf.get_delta_for_epsilon(epsilon))

  def test_delta_for_epsilon_not_sorted(self):
    pmf = self._create_pmf(discretization=0.1, dense=True, infinity_mass=0)
    epsilon = [2.0, 3.0, 1.0]  # not sorted

    with self.assertRaisesRegex(
        ValueError,
        'Epsilons in get_delta_for_epsilon must be sorted in ascending order'):
      pmf.get_delta_for_epsilon(epsilon)

  @parameterized.parameters(False, True)
  def test_get_delta_for_epsilon_for_composed_pld(self, dense):
    discretization = 0.1
    infinity_mass1, lower_loss1, probs1 = 0.1, -1, np.array(
        [0.2, 0.3, 0, 0.1, 0.3])
    infinity_mass2, lower_loss2, probs2 = 0.2, -2, np.array([0.1, 0, 0.4, 0.3])
    pmf1 = self._create_pmf(discretization, dense, infinity_mass1, lower_loss1,
                            probs1)
    pmf2 = self._create_pmf(discretization, dense, infinity_mass2, lower_loss2,
                            probs2)
    pmf_composed = pmf1.compose(pmf2)
    for epsilon in np.linspace(-10, 10, num=100):
      delta1 = pmf1.get_delta_for_epsilon_for_composed_pld(pmf2, epsilon)
      delta2 = pmf_composed.get_delta_for_epsilon(epsilon)
      self.assertAlmostEqual(delta1, delta2, msg=f'{epsilon}')

  @parameterized.parameters(False, True)
  def test_epsilon_for_delta(self, dense):
    discretization = 0.1
    lower_loss = -1  # loss_value
    probs = np.array([0.2, 0.3, 0, 0.4])  # probs for losses -0.1, 0, 0.1, 0.2
    infinity_mass = 0.1
    pmf = self._create_pmf(discretization, dense, infinity_mass, lower_loss,
                           probs)
    self.assertEqual(np.inf, pmf.get_epsilon_for_delta(0.05))  # <infinity_mass
    self.assertAlmostEqual(0.2, pmf.get_epsilon_for_delta(0.1))  # infinity_mass

    epsilon = 0.15
    delta = pmf.get_delta_for_epsilon(epsilon)
    self.assertAlmostEqual(epsilon, pmf.get_epsilon_for_delta(delta))

    self.assertAlmostEqual(np.inf, pmf.get_epsilon_for_delta(0))

  @parameterized.product(
      (
          {
              'tail_mass_truncation': 0,
              'expected_lower_loss': -3,
              'expected_probs': np.array([0.06, 0.31, 0.35]),
              'expected_truncated_to_inf_mass': 0
          },
          {
              'tail_mass_truncation':
                  0.1,  # no truncation 0.1/2 < min(0.06, 0.35)
              'expected_lower_loss': -3,
              'expected_probs': np.array([0.06, 0.31, 0.35]),
              'expected_truncated_to_inf_mass': 0
          },
          {
              'tail_mass_truncation': 0.15,  # truncation from left.
              'expected_lower_loss': -2,
              'expected_probs': np.array([0.37, 0.35]),
              'expected_truncated_to_inf_mass': 0
          },
          {
              'tail_mass_truncation': 0.72,  # truncation from both sides.
              'expected_lower_loss': -2,
              'expected_probs': np.array([0.37]),
              'expected_truncated_to_inf_mass':
                  0.35  # last element goes to inf.
          },
      ),
      dense=(False, True),
  )
  def test_compose_pessimistic(self, tail_mass_truncation, expected_lower_loss,
                               expected_probs, expected_truncated_to_inf_mass,
                               dense):
    discretization = 0.1
    pmf1 = self._create_pmf(
        discretization,
        lower_loss=-1,
        probs=np.array([0.2, 0.7]),
        infinity_mass=0.1,
        dense=dense)
    pmf2 = self._create_pmf(
        discretization,
        lower_loss=-2,
        probs=np.array([0.3, 0.5]),
        infinity_mass=0.2,
        dense=dense)
    pmf = pmf1.compose(pmf2, tail_mass_truncation)

    self.assertEqual(discretization, pmf._discretization)
    if dense:
      self._check_dense_probs(pmf, expected_lower_loss, expected_probs)
    else:
      self._check_sparse_probs(pmf, expected_lower_loss, expected_probs)

    expected_inf_mass = 1 - (1 - 0.1) * (1 -
                                         0.2) + expected_truncated_to_inf_mass
    self.assertAlmostEqual(expected_inf_mass, pmf._infinity_mass)

  @parameterized.product(
      (
          {
              'tail_mass_truncation': 0,
              'expected_lower_loss': -3,
              'expected_probs': np.array([0.06, 0.31, 0.35]),
              'expected_truncated_to_inf_mass': 0,
              'pessimistic_estimate': True
          },
          {
              'tail_mass_truncation': 0,
              'expected_lower_loss': -3,
              'expected_probs': np.array([0.06, 0.31, 0.35]),
              'expected_truncated_to_inf_mass': 0,
              'pessimistic_estimate': False
          },
          {
              'tail_mass_truncation':
                  0.1,  # no truncation 0.1/2 < min(0.06, 0.35)
              'expected_lower_loss': -3,
              'expected_probs': np.array([0.06, 0.31, 0.35]),
              'expected_truncated_to_inf_mass': 0,
              'pessimistic_estimate': True
          },
          {
              'tail_mass_truncation':
                  0.1,  # no truncation 0.1/2 < min(0.06, 0.35)
              'expected_lower_loss': -3,
              'expected_probs': np.array([0.06, 0.31, 0.35]),
              'expected_truncated_to_inf_mass': 0,
              'pessimistic_estimate': False
          },
          {
              'tail_mass_truncation': 0.15,  # truncation the left tail.
              'expected_lower_loss': -2,
              'expected_probs': np.array([0.37, 0.35]),
              'expected_truncated_to_inf_mass': 0,
              'pessimistic_estimate': True
          },
          {
              'tail_mass_truncation': 0.15,  # truncation the left tail.
              'expected_lower_loss': -2,
              'expected_probs': np.array([0.31, 0.35]),
              'expected_truncated_to_inf_mass': 0,
              'pessimistic_estimate': False
          },
          {
              'tail_mass_truncation': 0.72,  # truncation both tails.
              'expected_lower_loss': -2,
              'expected_probs': np.array([0.37]),
              'expected_truncated_to_inf_mass': 0.35,
              'pessimistic_estimate': True
          },
          {
              'tail_mass_truncation': 0.72,  # truncation both tails.
              'expected_lower_loss': -2,
              'expected_probs': np.array([0.66]),
              'expected_truncated_to_inf_mass': 0,
              'pessimistic_estimate': False
          },
      ),
      dense=(False, True))
  def test_compose(self, tail_mass_truncation: float, expected_lower_loss: int,
                   expected_probs: np.ndarray,
                   expected_truncated_to_inf_mass: float,
                   pessimistic_estimate: bool, dense: bool):
    discretization = 0.1
    pmf1 = self._create_pmf(
        discretization,
        lower_loss=-1,
        probs=np.array([0.2, 0.7]),
        infinity_mass=0.1,
        dense=dense,
        pessimistic_estimate=pessimistic_estimate)
    pmf2 = self._create_pmf(
        discretization,
        lower_loss=-2,
        probs=np.array([0.3, 0.5]),
        infinity_mass=0.2,
        dense=dense,
        pessimistic_estimate=pessimistic_estimate)
    pmf = pmf1.compose(pmf2, tail_mass_truncation)

    self.assertEqual(discretization, pmf._discretization)
    if dense:
      self._check_dense_probs(pmf, expected_lower_loss, expected_probs)
    else:
      self._check_sparse_probs(pmf, expected_lower_loss, expected_probs)

    expected_inf_mass = 1 - (1 - 0.1) * (1 -
                                         0.2) + expected_truncated_to_inf_mass
    self.assertAlmostEqual(expected_inf_mass, pmf._infinity_mass)

  @parameterized.parameters(False, True)
  def test_compose_different_discretization(self, dense: bool):
    pmf1 = self._create_pmf(discretization=0.1, dense=dense)
    pmf2 = self._create_pmf(discretization=0.2, dense=dense)

    with self.assertRaisesRegex(
        ValueError, 'Discretization intervals are different: 0.1 != 0.2'):
      pmf1.compose(pmf2)
    with self.assertRaisesRegex(
        ValueError, 'Discretization intervals are different: 0.1 != 0.2'):
      pmf1.get_delta_for_epsilon_for_composed_pld(pmf2, 1)

  @parameterized.parameters(False, True)
  def test_compose_different_estimation(self, dense: bool):
    pmf1 = self._create_pmf(
        discretization=0.1, pessimistic_estimate=True, dense=dense)
    pmf2 = self._create_pmf(
        discretization=0.1, pessimistic_estimate=False, dense=dense)

    with self.assertRaisesRegex(ValueError, 'Estimation types are different'):
      pmf1.compose(pmf2)
    with self.assertRaisesRegex(ValueError, 'Estimation types are different'):
      pmf1.get_delta_for_epsilon_for_composed_pld(pmf2, 1)

  @parameterized.parameters(
      {
          'num_times': 2,
          'tail_mass_truncation': 0,
          'expected_lower_loss': -2,
          'expected_probs': np.array([0.04, 0.28, 0.49]),
      }, {
          'num_times': 5,
          'tail_mass_truncation': 0,
          'expected_lower_loss': -5,
          'expected_probs':
              np.array([0.00032, 0.0056, 0.0392, 0.1372, 0.2401, 0.16807]),
      }, {
          'num_times': 2,
          'tail_mass_truncation': 0.1,
          'expected_lower_loss': -1,
          'expected_probs': np.array([0.32, 0.49]),
      }, {
          'num_times': 5,
          'tail_mass_truncation': 0.01,  # truncation left tail.
          'expected_lower_loss': -4,
          'expected_probs':
              np.array([0.00032 + 0.0056, 0.0392, 0.1372, 0.2401, 0.16807]),
      },
  )
  def test_self_compose_sparse(self, num_times, tail_mass_truncation,
                               expected_lower_loss, expected_probs):
    discretization = 0.1
    infinity_mass = 0.1
    pmf_input = self._create_pmf(
        discretization,
        lower_loss=-1,
        probs=np.array([0.2, 0.7]),
        infinity_mass=infinity_mass,
        dense=False)
    pmf_result = pmf_input.self_compose(num_times, tail_mass_truncation)

    self.assertEqual(discretization, pmf_result._discretization)
    self.assertAlmostEqual(1 - (1 - infinity_mass)**num_times,
                           pmf_result._infinity_mass)
    self._check_sparse_probs(pmf_result, expected_lower_loss, expected_probs)

  @parameterized.parameters(
      {
          'num_times': 2,
          'tail_mass_truncation': 0,
          'expected_lower_loss': -2,
          'expected_probs': np.array([0.04, 0.28, 0.49]),
      }, {
          'num_times': 5,
          'tail_mass_truncation': 0,
          'expected_lower_loss': -5,
          'expected_probs':
              np.array([0.00032, 0.0056, 0.0392, 0.1372, 0.2401, 0.16807]),
      },
  )
  def test_self_compose_dense_probs(self, num_times, tail_mass_truncation,
                                    expected_lower_loss, expected_probs):
    discretization = 0.1
    infinity_mass = 0.1
    pmf_input = self._create_pmf(
        discretization,
        lower_loss=-1,
        probs=np.array([0.2, 0.7]),
        infinity_mass=infinity_mass,
        dense=True)
    pmf_result = pmf_input.self_compose(num_times, tail_mass_truncation)
    self.assertEqual(discretization, pmf_result._discretization)
    self.assertAlmostEqual(1 - (1 - infinity_mass)**num_times,
                           pmf_result._infinity_mass)
    self._check_dense_probs(pmf_result, expected_lower_loss, expected_probs)

  def test_self_compose_dense_truncation(self):
    discretization = 0.1
    infinity_mass = 0.0001
    num_times = 50
    tail_mass_truncation = 1e-2
    pmf_input = self._create_pmf(
        discretization,
        lower_loss=-1,
        probs=np.array([0.2, 0.7]),
        infinity_mass=infinity_mass,
        dense=True)
    pmf_result = pmf_input.self_compose(num_times, tail_mass_truncation)

    self.assertEqual(discretization, pmf_result._discretization)
    self.assertLessEqual(
        pmf_result._infinity_mass,
        tail_mass_truncation - np.expm1(num_times * np.log1p(-infinity_mass)))

  @parameterized.parameters((1, True), (100, True), (1000, True), (1001, False))
  def test_pmf_creation(self, num_points: int, is_sparse: bool):
    probs = np.ones(num_points) / num_points
    lower_loss = -1
    loss_probs = common.list_to_dictionary(probs, lower_loss)
    discretization = 0.01
    infinity_mass = 0
    pessimistic_estimate = True
    pmf = pld_pmf.create_pmf(loss_probs, discretization, infinity_mass,
                             pessimistic_estimate)

    if is_sparse:
      self.assertIsInstance(pmf, pld_pmf.SparsePLDPmf)
    else:
      self.assertIsInstance(pmf, pld_pmf.DensePLDPmf)

    self.assertEqual(num_points, pmf.size)

  @parameterized.named_parameters(
      dict(testcase_name='empty',
           discretization=0.1,
           rounded_epsilons=np.array([]),
           deltas=np.array([]),
           error_message='unequal or zero'),
      dict(testcase_name='unequal_length',
           discretization=0.1,
           rounded_epsilons=np.array([-1, 1]),
           deltas=np.array([0.5]),
           error_message='unequal or zero'),
      dict(testcase_name='non_sorted_epsilons',
           discretization=0.1,
           rounded_epsilons=np.array([1, -1]),
           deltas=np.array([0, 0.5]),
           error_message='not in strictly increasing order'),
      dict(testcase_name='non_monotone_deltas',
           discretization=0.1,
           rounded_epsilons=np.array([-1, 0, 1]),
           deltas=np.array([0.5, 0, 0.1]),
           error_message='not in non-increasing order'),
      dict(testcase_name='out_of_range_deltas_1',
           discretization=0.1,
           rounded_epsilons=np.array([-1, 0, 1]),
           deltas=np.array([1.1, 0.5, 0.1]),
           error_message='deltas are not between 0 and 1'),
      dict(testcase_name='out_of_range_deltas_2',
           discretization=0.1,
           rounded_epsilons=np.array([-1, 0, 1]),
           deltas=np.array([0.9, 0.5, -0.1]),
           error_message='deltas are not between 0 and 1'),
      )
  def test_connect_dots_pmf_value_errors(
      self, discretization, rounded_epsilons, deltas, error_message):
    with self.assertRaisesRegex(ValueError, error_message):
      pld_pmf.create_pmf_pessimistic_connect_dots(discretization,
                                                  rounded_epsilons,
                                                  deltas)

  @parameterized.named_parameters(
      dict(testcase_name='example1',
           discretization=0.1,
           rounded_epsilons=np.array([0, 1]),
           deltas=np.array([0.4, 0.1])))
  def test_connect_dots_pmf_non_negative(self, discretization,
                                         rounded_epsilons, deltas):
    pmf = pld_pmf.create_pmf_pessimistic_connect_dots(
        discretization, rounded_epsilons, deltas)
    for k, v in pmf._loss_probs.items():
      self.assertGreaterEqual(v, 0, msg=f'Probability for loss={k} is {v} < 0')
    self.assertGreaterEqual(
        pmf._infinity_mass, 0,
        msg=f'Probability for +inf is {pmf._infinity_mass} < 0')

  @parameterized.named_parameters(
      dict(testcase_name='trivial',
           # mu_upper = [1]
           # mu_lower = [1]
           discretization=0.1,
           rounded_epsilons=np.array([0]),
           deltas=np.array([0.0]),
           expected_loss_probs={0: 1.0},
           expected_infinity_mass=0.0),
      dict(testcase_name='rr_eps=0.1',
           # mu_upper = [1/(1+e^{-0.1}), 1/(1+e^{0.1})]
           # mu_lower = [1/(1+e^{0.1}), 1/(1+e^{-0.1})]
           discretization=0.1,
           rounded_epsilons=np.array([-1, 1]),
           deltas=np.array([1 - np.exp(-0.1), 0.0]),
           expected_loss_probs={-1: 1/(1 + np.exp(0.1)),
                                1: 1/(1 + np.exp(-0.1))},
           expected_infinity_mass=0.0),
      dict(testcase_name='rr_eps=0.1_abort_w_p_0.5',
           # mu_upper = [0.5/(1+e^{-0.1}), 0.5, 0.5/(1+e^{0.1})]
           # mu_lower = [0.5/(1+e^{0.1}), 0.5, 0.5/(1+e^{-0.1})]
           discretization=0.1,
           rounded_epsilons=np.array([-1, 0, 1]),
           deltas=np.array([1 - np.exp(-0.1),
                            0.5*(np.exp(0.1)-1)/(np.exp(0.1)+1),
                            0.0]),
           expected_loss_probs={-1: 0.5/(1 + np.exp(0.1)),
                                0: 0.5,
                                1: 0.5/(1 + np.exp(-0.1))},
           expected_infinity_mass=0.0),
      dict(testcase_name='rr_eps=0.1_delta=0.5',
           # mu_upper = [0.5, 0.5/(1+e^{-0.1}), 0.5/(1+e^{0.1}), 0.0]
           # mu_lower = [0.0, 0.5/(1+e^{0.1}), 0.5/(1+e^{-0.1}), 0.5]
           discretization=0.1,
           rounded_epsilons=np.array([-1, 0, 1]),
           deltas=np.array([1 - 0.5*np.exp(-0.1),
                            0.5 + 0.5*(np.exp(0.1)-1)/(np.exp(0.1)+1),
                            0.5]),
           expected_loss_probs={-1: 0.5/(1 + np.exp(0.1)),
                                0: 0.0,
                                1: 0.5/(1 + np.exp(-0.1))},
           expected_infinity_mass=0.5),
      )
  def test_connect_dots_pmf_creation(
      self, discretization, rounded_epsilons, deltas,
      expected_loss_probs, expected_infinity_mass):
    pmf = pld_pmf.create_pmf_pessimistic_connect_dots(discretization,
                                                      rounded_epsilons,
                                                      deltas)
    self._check_sparse_pld_pmf_equal(pmf,
                                     discretization,
                                     expected_infinity_mass,
                                     expected_loss_probs,
                                     pessimistic_estimate=True)

  @parameterized.named_parameters(
      dict(testcase_name='epsilon_empty',
           discretization=0.1,
           rounded_epsilon_lower=1,
           rounded_epsilon_upper=0,
           deltas=np.array([]),
           error_message='is smaller than'),
      dict(testcase_name='unequal_length',
           discretization=0.1,
           rounded_epsilon_lower=0,
           rounded_epsilon_upper=1,
           deltas=np.array([0.5]),
           error_message='Length of deltas is not equal to number of epsilons'),
      dict(testcase_name='non_monotone_deltas',
           discretization=0.1,
           rounded_epsilon_lower=-1,
           rounded_epsilon_upper=1,
           deltas=np.array([0.5, 0, 0.1]),
           error_message='not in non-increasing order'),
      dict(testcase_name='out_of_range_deltas_1',
           discretization=0.1,
           rounded_epsilon_lower=-1,
           rounded_epsilon_upper=1,
           deltas=np.array([1.1, 0.5, 0.1]),
           error_message='deltas are not between 0 and 1'),
      dict(testcase_name='out_of_range_deltas_2',
           discretization=0.1,
           rounded_epsilon_lower=-1,
           rounded_epsilon_upper=1,
           deltas=np.array([0.9, 0.5, -0.1]),
           error_message='deltas are not between 0 and 1'),
      )
  def test_connect_dots_pmf_fixed_gap_value_errors(
      self, discretization, rounded_epsilon_lower, rounded_epsilon_upper,
      deltas, error_message):
    with self.assertRaisesRegex(ValueError, error_message):
      pld_pmf.create_pmf_pessimistic_connect_dots_fixed_gap(
          discretization, rounded_epsilon_lower, rounded_epsilon_upper, deltas)

  @parameterized.named_parameters(
      dict(testcase_name='example1',
           discretization=0.1,
           rounded_epsilon_lower=0,
           rounded_epsilon_upper=1,
           deltas=np.array([0.4, 0.1])))
  def test_connect_dots_pmf_fixed_gap_non_negative(
      self, discretization, rounded_epsilon_lower, rounded_epsilon_upper,
      deltas):
    pmf = pld_pmf.create_pmf_pessimistic_connect_dots_fixed_gap(
        discretization, rounded_epsilon_lower, rounded_epsilon_upper, deltas)
    for k, v in pmf._loss_probs.items():
      self.assertGreaterEqual(v, 0, msg=f'Probability for loss={k} is {v} < 0')
    self.assertGreaterEqual(
        pmf._infinity_mass, 0,
        msg=f'Probability for +inf is {pmf._infinity_mass} < 0')

  @parameterized.named_parameters(
      dict(testcase_name='trivial',
           # mu_upper = [1]
           # mu_lower = [1]
           discretization=0.1,
           rounded_epsilon_lower=0,
           rounded_epsilon_upper=0,
           deltas=np.array([0.0]),
           expected_loss_probs={0: 1.0},
           expected_infinity_mass=0.0),
      dict(testcase_name='rr_eps=0.1',
           # mu_upper = [1/(1+e^{-0.1}), 1/(1+e^{0.1})]
           # mu_lower = [1/(1+e^{0.1}), 1/(1+e^{-0.1})]
           discretization=0.1,
           rounded_epsilon_lower=-1,
           rounded_epsilon_upper=1,
           deltas=np.array([1 - np.exp(-0.1),
                            (1 - np.exp(-0.1))/(1 + np.exp(-0.1)),
                            0.0]),
           expected_loss_probs={-1: 1/(1 + np.exp(0.1)),
                                1: 1/(1 + np.exp(-0.1))},
           expected_infinity_mass=0.0),
      dict(testcase_name='rr_eps=0.1_abort_w_p_0.5',
           # mu_upper = [0.5/(1+e^{-0.1}), 0.5, 0.5/(1+e^{0.1})]
           # mu_lower = [0.5/(1+e^{0.1}), 0.5, 0.5/(1+e^{-0.1})]
           discretization=0.1,
           rounded_epsilon_lower=-1,
           rounded_epsilon_upper=1,
           deltas=np.array([1 - np.exp(-0.1),
                            0.5*(np.exp(0.1)-1)/(np.exp(0.1)+1),
                            0.0]),
           expected_loss_probs={-1: 0.5/(1 + np.exp(0.1)),
                                0: 0.5,
                                1: 0.5/(1 + np.exp(-0.1))},
           expected_infinity_mass=0.0),
      dict(testcase_name='rr_eps=0.1_delta=0.5',
           # mu_upper = [0.5, 0.5/(1+e^{-0.1}), 0.5/(1+e^{0.1}), 0.0]
           # mu_lower = [0.0, 0.5/(1+e^{0.1}), 0.5/(1+e^{-0.1}), 0.5]
           discretization=0.1,
           rounded_epsilon_lower=-1,
           rounded_epsilon_upper=1,
           deltas=np.array([1 - 0.5*np.exp(-0.1),
                            0.5 + 0.5*(np.exp(0.1)-1)/(np.exp(0.1)+1),
                            0.5]),
           expected_loss_probs={-1: 0.5/(1 + np.exp(0.1)),
                                0: 0.0,
                                1: 0.5/(1 + np.exp(-0.1))},
           expected_infinity_mass=0.5),
      )
  def test_connect_dots_fixed_gap_pmf_creation(
      self, discretization, rounded_epsilon_lower, rounded_epsilon_upper,
      deltas, expected_loss_probs, expected_infinity_mass):
    pmf = pld_pmf.create_pmf_pessimistic_connect_dots_fixed_gap(
        discretization, rounded_epsilon_lower, rounded_epsilon_upper, deltas)
    self._check_sparse_pld_pmf_equal(pmf,
                                     discretization,
                                     expected_infinity_mass,
                                     expected_loss_probs,
                                     pessimistic_estimate=True)

  @parameterized.parameters((1, 100, True), (10, 100, True), (1, 1001, False),
                            (10, 101, False), (1001, 1, False),
                            (1001, 1001, False))
  def test_compose_pmfs(self, num_points1, num_points2, is_sparse):
    lower_loss = -1
    discretization = 0.01
    infinity_mass = 0
    pessimistic_estimate = True
    probs1 = np.ones(num_points1) / num_points1
    loss_probs1 = common.list_to_dictionary(probs1, lower_loss)
    pmf1 = pld_pmf.create_pmf(loss_probs1, discretization, infinity_mass,
                              pessimistic_estimate)

    probs2 = np.ones(num_points2) / num_points2
    loss_probs2 = common.list_to_dictionary(probs2, lower_loss)
    pmf2 = pld_pmf.create_pmf(loss_probs2, discretization, infinity_mass,
                              pessimistic_estimate)

    pmf = pld_pmf.compose_pmfs(pmf1, pmf2)

    if is_sparse:
      self.assertIsInstance(pmf, pld_pmf.SparsePLDPmf)
    else:
      self.assertIsInstance(pmf, pld_pmf.DensePLDPmf)


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Probability mass function for privacy loss distributions.

This file implements work the privacy loss distribution (PLD) probability mass
functions (PMF)and its basic functionalities. Please refer to the
supplementary material below for more details:
../../common_docs/Privacy_Loss_Distributions.pdf
"""

import abc
import itertools
import math
import numbers
from typing import Iterable, List, Mapping, Sequence, Tuple, Union

import numpy as np
import numpy.typing
from scipy import signal

from dp_accounting.pld import common

ArrayLike = Union[np.ndarray, List[float]]
_MAX_PMF_SPARSE_SIZE = 1000


def _get_delta_for_epsilon(infinity_mass: float,
                           losses: Sequence[float],
                           probs: Sequence[float],
                           epsilon: float) -> float:
  """Computes the epsilon-hockey stick divergence.

  Args:
    infinity_mass: The probability of the infinite loss.
    losses: The privacy losses, assumed to be sorted in ascending order.
    probs: The probabilities corresponding to losses.
    epsilon: The epsilon in the epsilon-hockey stick divergence.

  Returns:
    The epsilon-hockey stick divergence.
  """
  # delta is inf_mass + sum_{loss} max(0, 1 - exp(epsilon - loss)) * prob
  losses = np.asarray(losses)
  probs = np.asarray(probs)
  indices = losses > epsilon
  return (
      infinity_mass +
      np.dot(-np.expm1(epsilon - losses[indices]), probs[indices])
  )


def _get_delta_for_epsilon_vectorized(infinity_mass: float,
                                      losses: Sequence[float],
                                      probs: Sequence[float],
                                      epsilons: Sequence[float]) -> np.ndarray:
  """Computes the epsilon-hockey stick divergence for multiple epsilons.

  Args:
    infinity_mass: the probability of the infinite loss.
    losses: privacy losses, assumed to be sorted in ascending order.
    probs: probabilities corresponding to losses.
    epsilons: epsilons in the epsilon-hockey stick divergence, assumed to be
      sorted in ascending order.

  Returns:
    The list of epsilon-hockey stick divergences for epsilons.
  """
  are_epsilons_sorted = np.all(np.diff(epsilons) >= 0)
  if not are_epsilons_sorted:
    raise ValueError(
        'Epsilons in get_delta_for_epsilon must be sorted in ascending order')

  # For each epsilon:
  # delta = sum_{o} [mu_upper(o) - e^{epsilon} * mu_lower(o)]_+ (for more
  # details look for PLDPmf class docstring below).
  # It can be rewritten as
  # delta(epsilon_i) = inf_mass +
  #           sum((1-exp(eps - loss_j))*prob_j if loss_j >= epsilon_i) =
  # inf_mass + sum(prob_j-exp(eps)*prob_j/exp(loss_j) if loss_j >= epsilon_i) =
  # inf_mass + sum(prob_j, loss_j >= epsilon_i) -
  # exp(eps)*sum(prob_j/exp(loss_j), loss_j >= epsilon_i).
  #
  # Denote sums in the last formula as mu_upper_mass, mu_lower_mass. We can
  # compute them gradually, which makes computation efficiently for multiple
  # epsilons.
  deltas = np.zeros_like(epsilons, dtype=np.float64)
  mu_upper_mass, mu_lower_mass = infinity_mass, 0
  i = len(probs) - 1
  j = len(epsilons) - 1

  while j >= 0:
    if np.isposinf(epsilons[j]):
      deltas[j] = infinity_mass
      j -= 1
    elif i < 0 or losses[i] <= epsilons[j]:
      deltas[j] = mu_upper_mass - np.exp(epsilons[j]) * mu_lower_mass
      j -= 1
    else:
      mu_upper_mass += probs[i]
      mu_lower_mass += probs[i] * np.exp(-losses[i])
      i -= 1

  return deltas


def _get_epsilon_for_delta(infinity_mass: float,
                           reversed_losses: Iterable[float],
                           probs: Iterable[float], delta: float) -> float:
  """Computes epsilon for which hockey stick divergence is at most delta.

  Args:
    infinity_mass: the probability of the infinite loss.
    reversed_losses: privacy losses, assumed to be sorted in descending order.
    probs: probabilities corresponding to losses.
    delta: the target epsilon-hockey stick divergence..

  Returns:
     The smallest epsilon such that the epsilon-hockey stick divergence is at
     most delta. When no such finite epsilon exists, return math.inf.
  """
  if infinity_mass > delta:
    return math.inf

  mass_upper, mass_lower = infinity_mass, 0

  for loss, prob in zip(reversed_losses, probs):
    if (mass_upper > delta and mass_lower > 0 and math.log(
        (mass_upper - delta) / mass_lower) >= loss):
      # Epsilon is greater than or equal to loss.
      break

    mass_upper += prob
    mass_lower += math.exp(-loss) * prob

    if mass_upper >= delta and mass_lower == 0:
      # This only occurs when loss is very large, which results in exp(-loss)
      # being treated as zero.
      return max(0, loss)

  if mass_upper <= mass_lower + delta:
    return 0
  return math.log((mass_upper - delta) / mass_lower)


def _truncate_tails(probs: ArrayLike, tail_mass_truncation: float,
                    pessimistic_estimate: bool) -> Tuple[int, ArrayLike, float]:
  """Truncates an array from both sides by not more than tail_mass_truncation.

  It truncates the maximum prefix and suffix from probs, each of which have
  sum <= tail_mass_truncation/2.

  Args:
    probs: array to truncate.
    tail_mass_truncation: an upper bound on the tails of the probability mass of
      the PMF that might be truncated.
    pessimistic_estimate: if true then the left truncated sum is added to 0th
      element of the truncated array and the right truncated returned as it goes
      to infinity. If false then the right truncated sum is added to the last of
      the truncated array and the left truncated sum is discarded.

  Returns:
    Tuple of (size of truncated prefix, truncated array, mass that goes to
    infinity).
  """
  if tail_mass_truncation == 0:
    return 0, probs, 0

  def _find_prefix_to_truncate(arr: np.ndarray, threshold: float) -> int:
    # Find the max size of array prefix, with the sum of elements less than
    # threshold.
    s = 0
    for i, val in enumerate(arr):
      s += val
      if s > threshold:
        return i
    return len(arr)

  left_idx = _find_prefix_to_truncate(probs, tail_mass_truncation / 2)
  right_idx = len(probs) - _find_prefix_to_truncate(
      np.flip(probs), tail_mass_truncation / 2)
  # Be sure that left_idx < right_idx. left_idx >= right_idx might be when
  # tail_mass_truncation is too large or if probs has too small mass
  # (i.e. if a few truncations were operated on it already).
  right_idx = max(right_idx, left_idx + 1)

  left_mass = np.sum(probs[:left_idx])
  right_mass = np.sum(probs[right_idx:])

  truncated_probs = probs[left_idx:right_idx]
  if pessimistic_estimate:
    # put truncated the left mass to the 0th element.
    truncated_probs[0] += left_mass
    return left_idx, truncated_probs, right_mass
  # This is rounding to left case. Put truncated the right mass to the last
  # element.
  truncated_probs[-1] += right_mass
  return left_idx, truncated_probs, 0


class PLDPmf(abc.ABC):
  """Base class for probability mass functions for privacy loss distributions.

  The privacy loss distribution (PLD) of two discrete distributions, the upper
  distribution mu_upper and the lower distribution mu_lower, is defined as a
  distribution on real numbers generated by first sampling an outcome o
  according to mu_upper and then outputting the privacy loss
  ln(mu_upper(o) / mu_lower(o)) where mu_lower(o) and mu_upper(o) are the
  probability masses of o in mu_lower and mu_upper respectively. This class
  allows one to create and manipulate privacy loss distributions.

  PLD allows one to (approximately) compute the epsilon-hockey stick divergence
  between mu_upper and mu_lower, which is defined as
  sum_{o} [mu_upper(o) - e^{epsilon} * mu_lower(o)]_+. This quantity in turn
  governs the parameter delta of (eps, delta)-differential privacy of the
  corresponding protocol. (See Observation 1 in the supplementary material.)

  The above definitions extend to continuous distributions. The PLD of two
  continuous distributions mu_upper and mu_lower is defined as a distribution on
  real numbers generated by first sampling an outcome o according to mu_upper
  and then outputting the privacy loss ln(f_{mu_upper}(o) / f_{mu_lower}(o))
  where f_{mu_lower}(o) and f_{mu_upper}(o) are the probability density
  functions at o in mu_lower and mu_upper respectively. Moreover, for continuous
  distributions the epsilon-hockey stick divergence is defined as
  integral [f_{mu_upper}(o) - e^{epsilon} * f_{mu_lower}(o)]_+ do.
  """

  def __init__(self, discretization: float, infinity_mass: float,
               pessimistic_estimate: bool):
    self._discretization = discretization
    self._infinity_mass = infinity_mass
    self._pessimistic_estimate = pessimistic_estimate

  @property
  @abc.abstractmethod
  def size(self) -> int:
    """Returns number of points in discretization."""

  @abc.abstractmethod
  def compose(self,
              other: 'PLDPmf',
              tail_mass_truncation: float = 0) -> 'PLDPmf':
    """Computes a PMF resulting from composing two PMFs.

    Args:
      other: the privacy loss distribution PMF to be composed. The two must have
        the same discretization and pessimistic_estimate.
      tail_mass_truncation: an upper bound on the tails of the probability mass
        of the PMF that might be truncated.

    Returns:
      A PMF which is the result of convolving (composing) the two.
    """

  @abc.abstractmethod
  def self_compose(self,
                   num_times: int,
                   tail_mass_truncation: float = 0) -> 'PLDPmf':
    """Computes PMF resulting from repeated composing the PMF with itself.

    Args:
      num_times: the number of times to compose this PMF with itself.
      tail_mass_truncation: an upper bound on the tails of the probability mass
        of the PMF that might be truncated.

    Returns:
      A privacy loss distribution PMF which is the result of the composition.
    """

  @abc.abstractmethod
  def get_delta_for_epsilon(
      self, epsilon: Union[float, Sequence[float]]) -> Union[float, np.ndarray]:
    """Computes the epsilon-hockey stick divergence."""

  @abc.abstractmethod
  def get_epsilon_for_delta(self, delta: float) -> float:
    """Computes epsilon for which hockey stick divergence is at most delta."""

  @abc.abstractmethod
  def to_dense_pmf(self) -> 'DensePLDPmf':
    """Returns the dense PMF with data from 'self'."""

  @abc.abstractmethod
  def get_delta_for_epsilon_for_composed_pld(self, other: 'PLDPmf',
                                             epsilon: float) -> float:
    """Computes delta for 'epsilon' for the composiion of 'self' and 'other'."""

  def validate_composable(self, other: 'PLDPmf'):
    """Checks whether 'self' and 'other' can be composed."""
    if not isinstance(self, type(other)):
      raise ValueError(f'Only PMFs of the same type can be composed:'
                       f'{type(self).__name__} != {type(other).__name__}.')
    # pylint: disable=protected-access
    if self._discretization != other._discretization:
      raise ValueError(f'Discretization intervals are different: '
                       f'{self._discretization} != '
                       f'{other._discretization}.')
    if self._pessimistic_estimate != other._pessimistic_estimate:
      raise ValueError(f'Estimation types are different: '
                       f'{self._pessimistic_estimate} != '
                       f'{other._pessimistic_estimate}.')  # pylint: disable=protected-access
    # pylint: enable=protected-access


class DensePLDPmf(PLDPmf):
  """Class for dense probability mass function.

  It represents a discrete probability distribution on a grid of privacy losses.
  The grid contains numbers multiple of 'discretization', starting from
  lower_loss * discretization.
  """

  def __init__(self, discretization: float, lower_loss: int, probs: np.ndarray,
               infinity_mass: float, pessimistic_estimate: bool):
    super().__init__(discretization, infinity_mass, pessimistic_estimate)
    self._lower_loss = lower_loss
    self._probs = probs

  @property
  def size(self) -> int:
    return len(self._probs)

  def compose(self,
              other: 'DensePLDPmf',
              tail_mass_truncation: float = 0) -> 'DensePLDPmf':
    """Computes a PMF resulting from composing two PMFs. See base class."""
    self.validate_composable(other)

    # pylint: disable=protected-access
    lower_loss = self._lower_loss + other._lower_loss
    probs = signal.fftconvolve(self._probs, other._probs)
    infinity_mass = (self._infinity_mass + other._infinity_mass
                     - self._infinity_mass * other._infinity_mass)
    offset, probs, right_tail = _truncate_tails(probs, tail_mass_truncation,
                                                self._pessimistic_estimate)
    # pylint: enable=protected-access
    return DensePLDPmf(self._discretization, lower_loss + offset, probs,
                       infinity_mass + right_tail, self._pessimistic_estimate)

  def self_compose(self,
                   num_times: int,
                   tail_mass_truncation: float = 1e-15) -> 'DensePLDPmf':
    """See base class."""
    if num_times <= 0:
      raise ValueError(f'num_times should be >= 1, num_times={num_times}')
    lower_loss = self._lower_loss * num_times
    truncation_lower_bound, probs = common.self_convolve(
        self._probs, num_times, tail_mass_truncation)
    lower_loss += truncation_lower_bound
    probs = np.array(probs)
    # infinity mass after composition is given as
    # tail_mass_truncation + 1 - (1 - infinity_mass)**num_times
    # Below we use a numerically stable approach to compute the second term.
    inf_prob = (tail_mass_truncation
                - math.expm1(num_times * math.log1p(- self._infinity_mass)))
    return DensePLDPmf(self._discretization, lower_loss, probs,
                       inf_prob, self._pessimistic_estimate)

  def get_delta_for_epsilon(
      self, epsilon: Union[float, Sequence[float]]) -> Union[float, np.ndarray]:
    """Computes the epsilon-hockey stick divergence."""
    losses = (np.arange(self.size) + self._lower_loss) * self._discretization

    if isinstance(epsilon, numbers.Number):
      return _get_delta_for_epsilon(self._infinity_mass, losses,
                                    self._probs, epsilon)
    return _get_delta_for_epsilon_vectorized(self._infinity_mass, losses,
                                             self._probs, epsilon)

  def get_epsilon_for_delta(self, delta: float) -> float:
    """Computes epsilon for which hockey stick divergence is at most delta."""
    upper_loss = (self._lower_loss + len(self._probs) -
                  1) * self._discretization
    reversed_losses = itertools.count(upper_loss, -self._discretization)

    return _get_epsilon_for_delta(self._infinity_mass, reversed_losses,
                                  np.flip(self._probs), delta)

  def to_dense_pmf(self) -> 'DensePLDPmf':
    return self

  def get_delta_for_epsilon_for_composed_pld(self, other: PLDPmf,
                                             epsilon: float) -> float:
    other = other.to_dense_pmf()
    self.validate_composable(other)
    discretization = self._discretization
    # pylint: disable=protected-access
    self_loss = lambda index: (index + self._lower_loss) * discretization
    other_loss = lambda index: (index + other._lower_loss) * discretization

    self_probs, other_probs = self._probs, other._probs
    len_self, len_other = len(self_probs), len(other_probs)
    delta = (self._infinity_mass + other._infinity_mass
             - self._infinity_mass * other._infinity_mass)
    # pylint: enable=protected-access

    # Compute the hockey stick divergence using equation (2) in the
    # supplementary material. upper_mass represents summation in equation (3)
    # and lower_mass represents the summation in equation (4).

    if self_loss(len_self - 1) + other_loss(len_other - 1) <= epsilon:
      return delta

    i, j = 0, len_other - 1
    upper_mass = lower_mass = 0

    # This is summation by i,j, such that self_loss(i) + other_loss(j) >=
    # epsilon, and self_loss(i) + other_loss(j-1)< epsilon, as in the
    # equation(2).

    # If i is todo small then increase it.
    while self_loss(i) + other_loss(j) < epsilon:
      i += 1

    # Else if j is too large then decrease it.
    while j >= 0 and self_loss(i) + other_loss(j - 1) >= epsilon:
      upper_mass += other_probs[j]
      lower_mass += other_probs[j] * np.exp(-other_loss(j))
      j -= 1

    # Invariant:
    # self_loss(i) + other_loss(j-1) < epsilon <= self_loss(i) + other_loss(j)
    # Sum over all i, keeping this invariant.
    for i in range(i, len_self):
      if j >= 0:
        upper_mass += other_probs[j]
        lower_mass += other_probs[j] * np.exp(-other_loss(j))
      j -= 1
      delta += self_probs[i] * (
          upper_mass - np.exp(epsilon - self_loss(i)) * lower_mass)

    return delta


class SparsePLDPmf(PLDPmf):
  """Class for sparse probability mass function.

  It represents a discrete probability distribution on a grid of 1d losses with
  a dictionary. The grid contains numbers multiples of 'discretization'.
  """

  def __init__(self, loss_probs: Mapping[int, float], discretization: float,
               infinity_mass: float, pessimistic_estimate: bool):
    super().__init__(discretization, infinity_mass, pessimistic_estimate)
    self._loss_probs = loss_probs

  @property
  def size(self) -> int:
    return len(self._loss_probs)

  def compose(self,
              other: 'SparsePLDPmf',
              tail_mass_truncation: float = 0) -> 'SparsePLDPmf':
    """Computes a PMF resulting from composing two PMFs. See base class."""
    self.validate_composable(other)
    # Assumed small number of points, so simple quadratic algorithm is fine.
    convolution = {}
    # pylint: disable=protected-access
    for key1, value1 in self._loss_probs.items():
      for key2, value2 in other._loss_probs.items():
        key = key1 + key2
        convolution[key] = convolution.get(key, 0.0) + value1 * value2
    infinity_mass = (self._infinity_mass + other._infinity_mass
                     - self._infinity_mass * other._infinity_mass)
    # pylint: enable=protected-access
    # Do truncation.
    sorted_losses = sorted(convolution.keys())
    probs = [convolution[loss] for loss in sorted_losses]
    offset, probs, right_mass = _truncate_tails(probs, tail_mass_truncation,
                                                self._pessimistic_estimate)
    sorted_losses = sorted_losses[offset:offset + len(probs)]
    truncated_convolution = dict(zip(sorted_losses, probs))
    return SparsePLDPmf(truncated_convolution, self._discretization,
                        infinity_mass + right_mass, self._pessimistic_estimate)

  def self_compose(self,
                   num_times: int,
                   tail_mass_truncation: float = 1e-15) -> 'PLDPmf':
    """See base class."""
    if num_times <= 0:
      raise ValueError(f'num_times should be >= 1, num_times={num_times}')
    if num_times == 1:
      return self

    # Compute a rough upper bound overestimate, since from some power, the PMF
    # becomes dense and start growing linearly further. But in this case we
    # should definitely go to dense.
    max_result_size = self.size**num_times

    if max_result_size > _MAX_PMF_SPARSE_SIZE:
      # The size of composed PMF is too large for sparse. Convert to dense.
      return self.to_dense_pmf().self_compose(num_times, tail_mass_truncation)

    result = self
    for i in range(2, num_times + 1):
      # To truncate only on the last composition.
      mass_truncation = 0 if i != num_times else tail_mass_truncation
      result = result.compose(self, mass_truncation)

    return result

  def _get_losses_probs(self) -> Tuple[List[float], List[float]]:
    """Returns losses, sorted ascendingly and respective probabilities."""
    losses = sorted(list(self._loss_probs.keys()))
    probs = [self._loss_probs[loss] for loss in losses]
    losses = [loss * self._discretization for loss in losses]
    return losses, probs

  def get_delta_for_epsilon(
      self, epsilon: Union[float, Sequence[float]]) -> Union[float, np.ndarray]:
    """Computes the epsilon-hockey stick divergence."""
    losses, probs = self._get_losses_probs()
    if isinstance(epsilon, numbers.Number):
      return _get_delta_for_epsilon(self._infinity_mass, losses, probs, epsilon)

    return _get_delta_for_epsilon_vectorized(self._infinity_mass, losses,
                                             probs, epsilon)

  def get_epsilon_for_delta(self, delta: float) -> float:
    """Computes epsilon for which hockey stick divergence is at most delta."""
    losses, probs = self._get_losses_probs()
    return _get_epsilon_for_delta(self._infinity_mass, losses[::-1],
                                  probs[::-1], delta)

  def get_delta_for_epsilon_for_composed_pld(self, other: PLDPmf,
                                             epsilon: float) -> float:
    # If 'self' is sparse, then it is small, so it is not so expensive to
    # convert to dense. Let us convert it for simplicity for dense.
    return self.to_dense_pmf().get_delta_for_epsilon_for_composed_pld(
        other, epsilon)

  def to_dense_pmf(self) -> DensePLDPmf:
    """"Converts to dense PMF."""
    lower_loss, probs = common.dictionary_to_list(self._loss_probs)
    return DensePLDPmf(self._discretization, lower_loss, np.array(probs),
                       self._infinity_mass, self._pessimistic_estimate)


def create_pmf(loss_probs: Mapping[int, float], discretization: float,
               infinity_mass: float, pessimistic_estimate: bool) -> PLDPmf:
  """Creates PLDPmfs.

  It returns SparsePLDPmf if the size of loss_probs less than
   MAX_PMF_SPARSE_SIZE, otherwise DensePLDPmf.

  Args:
    loss_probs: probability mass function of the discretized privacy loss
      distribution.
    discretization: the interval length for which the values of the privacy loss
      distribution are discretized.
    infinity_mass: infinity_mass for privacy loss distribution.
    pessimistic_estimate: whether the rounding is done in such a way that the
      resulting epsilon-hockey stick divergence computation gives an upper
      estimate to the real value.

  Returns:
    Created PLDPmf.
  """
  if len(loss_probs) <= _MAX_PMF_SPARSE_SIZE:
    return SparsePLDPmf(loss_probs, discretization, infinity_mass,
                        pessimistic_estimate)

  lower_loss, probs = common.dictionary_to_list(loss_probs)
  probs = np.array(probs)
  return DensePLDPmf(discretization, lower_loss, probs, infinity_mass,
                     pessimistic_estimate)


def create_pmf_pessimistic_connect_dots(
    discretization: float,
    rounded_epsilons: numpy.typing.ArrayLike,  # dtype int, shape (n,)
    deltas: numpy.typing.ArrayLike,  # dtype float, shape (n,)
    ) -> PLDPmf:
  """Returns PLD probability mass function using pessimistic Connect-the-Dots.

  This method uses Algorithm 1 (PLD Discretization) from the following paper:
    Title: Connect the Dots: Tighter Discrete Approximations of Privacy Loss
           Distributions
    Authors: V. Doroshenko, B. Ghazi, P. Kamath, R. Kumar, P. Manurangsi
    Link: https://arxiv.org/abs/2207.04380

  Given a set of (finite) epsilon values that the PLD should be supported on
  (in addition to +infinity), the probability mass is computed as follows:
  Suppose desired epsilon values are [eps_1, ... , eps_n].
  Let eps_0 = -infinity for convenience.
  Let delta_i = eps_i-hockey stick divergence for the mechanism.
    (Note: delta_0 = 1)

  The probability mass on eps_i (1 <= i <= n-1) is given as:
    ((delta_i - delta_{i-1}) / (exp(eps_{i-1} - eps_i) - 1)
     + (delta_{i+1} - delta_i) / (exp(eps_{i+1} - eps_i) - 1))
  The probability mass on eps_n is given as:
    (delta_n - delta_{n-1}) / (exp(eps_{n-1} - eps_n) - 1)
  The probability mass on +infinity is simply delta_n.

  Args:
    discretization: The length of the discretization interval, such that all
      epsilon values in the support of the privacy loss distribution are integer
      multiples of it.
    rounded_epsilons: The desired support of the privacy loss distribution
      specified as a strictly increasing sequence of integer values. The support
      will be given by these values times discretization.
    deltas: The delta values corresponding to the epsilon values. These values
      must be in non-increasing order, due to the nature of hockey stick
      divergence.

  Returns:
    The pessimistic Connect-the-Dots privacy loss distribution supported on
    specified epsilons.

  Raises:
    ValueError: If any of the following hold:
      - rounded_epsilons and deltas do not have the same length, or if one of
        them is empty, or
      - if rounded_epsilons are not in strictly increasing order, or
      - if deltas are not in non-increasing order.
  """
  rounded_epsilons = np.asarray(rounded_epsilons, dtype=int)
  deltas = np.asarray(deltas, dtype=float)

  if (rounded_epsilons.size != deltas.size or rounded_epsilons.size == 0):
    raise ValueError('Length of rounded_epsilons and deltas are either unequal '
                     f'or zero: rounded_epsilons={rounded_epsilons}, '
                     f'deltas={deltas}.')

  # Notation: epsilons = [eps_1, ... , eps_n]
  # epsilon_diffs = [eps_2 - eps_1, ... , eps_n - eps_{n-1}]
  epsilon_diffs = np.diff(rounded_epsilons) * discretization
  if np.any(epsilon_diffs <= 0):
    raise ValueError('rounded_epsilons are not in strictly increasing order: '
                     f'rounded_epsilons={rounded_epsilons}.')

  # Notation: deltas = [delta_1, ... , delta_n]
  if np.any(deltas < 0) or np.any(deltas > 1):
    raise ValueError(f'deltas are not between 0 and 1 : deltas={deltas}.')
  # delta_diffs = [delta_2 - delta_1, ... , delta_n - delta_{n-1}]
  delta_diffs = np.diff(deltas)
  if np.any(delta_diffs > 0):
    raise ValueError(f'deltas are not in non-increasing order: '
                     f'deltas={deltas}.')

  # delta_diffs_scaled_v1 = [y_0, y_1, ..., y_{n-1}]
  # where y_i = (delta_{i+1} - delta_i) / (exp(eps_i - eps_{i+1}) - 1)
  # and   y_0 = (1 - delta_1)
  delta_diffs_scaled_v1 = np.append(1 - deltas[0],
                                    delta_diffs / np.expm1(- epsilon_diffs))

  # delta_diffs_scaled_v2 = [z_1, z_2, ..., z_{n-1}, z_n]
  # where z_i = (delta_{i+1} - delta_i) / (exp(eps_{i+1} - eps_i) - 1)
  # and   z_n = 0
  delta_diffs_scaled_v2 = np.append(delta_diffs / np.expm1(epsilon_diffs), 0.0)

  # PLD contains eps_i with probability mass y_{i-1} + z_i, and infinity with
  # probability mass delta_n. Enforce that probabilities are non-negative.
  probs = np.maximum(0, delta_diffs_scaled_v1 + delta_diffs_scaled_v2)

  return create_pmf(
      loss_probs=dict(zip(rounded_epsilons, probs)),
      discretization=discretization,
      infinity_mass=deltas[-1],
      pessimistic_estimate=True)


def create_pmf_pessimistic_connect_dots_fixed_gap(
    discretization: float,
    rounded_epsilon_lower: int,
    rounded_epsilon_upper: int,
    deltas: numpy.typing.ArrayLike,  # dtype float, shape (n,)
    ) -> PLDPmf:
  """Returns pessimistic Connect-the-Dots PLD PMF for epsilons with fixed gaps.

  This method uses Algorithm 1 (PLD Discretization) from the following paper:
    Title: Connect the Dots: Tighter Discrete Approximations of Privacy Loss
           Distributions
    Authors: V. Doroshenko, B. Ghazi, P. Kamath, R. Kumar, P. Manurangsi
    Link: https://arxiv.org/abs/2207.04380

  Given a set of (finite) epsilon values that the PLD should be supported on
  (in addition to +infinity), the probability mass is computed as follows:
  Suppose desired epsilon values are [eps_1, ... , eps_n], where
    eps_i - eps_{i-1} = discretization, for all i in {2, ... , n}.
  Let delta_i = eps_i-hockey stick divergence for the mechanism, and
  let d = discretization (for short).

  Case: n == 1
    The probability mass on eps_1 is given as (1 - delta_1), and
    the probability mass on +infinity is simply delta_1.

  Case: n >= 2
  The probability mass on eps_1 is given as
    (1 - delta_1) + (delta_2 - delta_1) / (exp(d) - 1)
  The probability mass on eps_i (2 <= i <= n-1) is given as:
    ((delta_{i+1} - delta_i) - exp(d) * (delta_i - delta_{i-1})) / (exp(d) - 1)
  The probability mass on eps_n is given as:
    (delta_n - delta_{n-1}) / (exp(-d) - 1)
  The probability mass on +infinity is simply delta_n.

  This method is a more numerically stable special case of
  create_pmf_pessimistic_connect_dots, where epsilons are contiguous integer
  multiples of discretization, and hence can be specified by simply the
  smallest and the largest epsilon values.

  Args:
    discretization: The length of the discretization interval, such that all
      epsilon values in the support of the privacy loss distribution are integer
      multiples of it.
    rounded_epsilon_lower: The smallest epsilon value divided by discretization.
    rounded_epsilon_upper: The largest epsilon value divided by discretization.
    deltas: The delta values corresponding to the epsilon values. These values
      must be in non-increasing order, due to the nature of hockey stick
      divergence.

  Returns:
    The pessimistic Connect-the-Dots privacy loss distribution supported on
    specified epsilons.

  Raises:
    ValueError: If any of the following hold:
      - rounded_epsilon_upper < rounded_epsilon_lower, or
      - length of deltas does not equal the number of epsilons, or
      - if deltas are not in non-increasing order.
  """
  deltas = np.asarray(deltas, dtype=np.float64)

  if rounded_epsilon_upper < rounded_epsilon_lower:
    raise ValueError(f'rounded_epsilon_upper={rounded_epsilon_upper} is smaller'
                     f' than rounded_epsilon_lower={rounded_epsilon_lower}.')
  if deltas.size != rounded_epsilon_upper - rounded_epsilon_lower + 1:
    raise ValueError('Length of deltas is not equal to number of epsilons: '
                     f'number of epsilons = '
                     f'{rounded_epsilon_upper - rounded_epsilon_lower + 1}, '
                     f'number of deltas = {deltas.size}.')
  if np.any(deltas < 0) or np.any(deltas > 1):
    raise ValueError(f'deltas are not between 0 and 1 : deltas={deltas}.')

  # Case of single epsilon in support:
  if rounded_epsilon_lower == rounded_epsilon_upper:
    return create_pmf(loss_probs={rounded_epsilon_lower: 1 - deltas[0]},
                      discretization=discretization,
                      infinity_mass=deltas[0],
                      pessimistic_estimate=True)

  # Notation: deltas = [delta_1, ... , delta_n]
  # delta_diffs = [delta_2 - delta_1, ... , delta_n - delta_{n-1}]
  delta_diffs = np.diff(deltas)
  if np.any(delta_diffs > 0):
    raise ValueError(f'deltas are not in non-increasing order: '
                     f'deltas={deltas}.')

  # probs = [p_1, p_2, ... , p_n]
  probs = np.zeros_like(deltas)

  # p_1 = 1 - delta_1 + (delta_2 - delta_1) / (exp(d) - 1),
  probs[0] = 1 - deltas[0] + delta_diffs[0] / math.expm1(discretization)

  # for all i such that 2 <= i <= n-1 : p_i =
  # ((delta_{i+1} - delta_i) - exp(d) * (delta_i - delta_{i-1})) / (exp(d) - 1)
  probs[1:-1] = ((delta_diffs[1:] - delta_diffs[:-1] * math.exp(discretization))
                 / math.expm1(discretization))

  # p_n = (delta_n - delta_{n-1}) / (exp(-d) - 1)
  probs[-1] = delta_diffs[-1] / math.expm1(-discretization)

  # Enforce that probabilities are non-negative.
  probs = np.maximum(0.0, probs)

  return create_pmf(
      loss_probs=dict(
          zip(range(rounded_epsilon_lower, rounded_epsilon_upper + 1), probs)),
      discretization=discretization,
      infinity_mass=deltas[-1],
      pessimistic_estimate=True)


def compose_pmfs(pmf1: PLDPmf,
                 pmf2: PLDPmf,
                 tail_mass_truncation: float = 0) -> PLDPmf:
  """Computes a PMF resulting from composing two PMFs.

  It returns SparsePLDPmf only if input PLDPmfs are SparsePLDPmf and the
  product of input pmfs sizes are less than MAX_PMF_SPARSE_SIZE.

  Args:
    pmf1: the privacy loss distribution PMF to be composed.
    pmf2: the privacy loss distribution PMF to be composed. The two must have
      the same discretization and pessimistic_estimate.
    tail_mass_truncation: an upper bound on the tails of the probability mass of
      the PMF that might be truncated.

  Returns:
    A PMF which is the result of convolving (composing) the two.
  """
  max_result_size = pmf1.size * pmf2.size
  if (isinstance(pmf1, SparsePLDPmf) and isinstance(pmf2, SparsePLDPmf) and
      max_result_size <= _MAX_PMF_SPARSE_SIZE):
    return pmf1.compose(pmf2, tail_mass_truncation)

  pmf1 = pmf1.to_dense_pmf()
  pmf2 = pmf2.to_dense_pmf()
  return pmf1.compose(pmf2, tail_mass_truncation)



# Copyright 2021, The TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for DpEventBuilder."""

from absl.testing import absltest
from dp_accounting import dp_event
from dp_accounting import dp_event_builder

_gaussian_event = dp_event.GaussianDpEvent(1.0)
_laplace_event = dp_event.LaplaceDpEvent(1.0)
_poisson_event = dp_event.PoissonSampledDpEvent(_gaussian_event, 0.1)
_self_composed_event = dp_event.SelfComposedDpEvent(_gaussian_event, 3)


class DpEventBuilderTest(absltest.TestCase):

  def test_no_op(self):
    builder = dp_event_builder.DpEventBuilder()
    self.assertEqual(dp_event.NoOpDpEvent(), builder.build())

  def test_single_gaussian(self):
    builder = dp_event_builder.DpEventBuilder()
    builder.compose(_gaussian_event)
    self.assertEqual(_gaussian_event, builder.build())

  def test_single_laplace(self):
    builder = dp_event_builder.DpEventBuilder()
    builder.compose(_laplace_event)
    self.assertEqual(_laplace_event, builder.build())

  def test_compose_no_op(self):
    builder = dp_event_builder.DpEventBuilder()
    builder.compose(dp_event.NoOpDpEvent())
    builder.compose(_gaussian_event)
    builder.compose(dp_event.NoOpDpEvent())
    self.assertEqual(_gaussian_event, builder.build())

  def test_compose_self(self):
    builder = dp_event_builder.DpEventBuilder()
    builder.compose(_gaussian_event)
    builder.compose(_gaussian_event, 2)
    self.assertEqual(_self_composed_event, builder.build())

  def test_compose_heterogenous(self):
    builder = dp_event_builder.DpEventBuilder()
    builder.compose(_poisson_event)
    builder.compose(_gaussian_event)
    builder.compose(_gaussian_event, 2)
    builder.compose(_poisson_event)
    expected_event = dp_event.ComposedDpEvent(
        [_poisson_event, _self_composed_event, _poisson_event])
    self.assertEqual(expected_event, builder.build())

  def test_compose_composed(self):
    builder = dp_event_builder.DpEventBuilder()
    composed_event = dp_event.ComposedDpEvent(
        [_gaussian_event, _poisson_event, _self_composed_event])
    builder.compose(_gaussian_event)
    builder.compose(composed_event)
    builder.compose(composed_event, 2)
    builder.compose(_poisson_event)
    builder.compose(_poisson_event)
    expected_event = dp_event.ComposedDpEvent([
        _gaussian_event,
        dp_event.SelfComposedDpEvent(composed_event, 3),
        dp_event.SelfComposedDpEvent(_poisson_event, 2)
    ])
    self.assertEqual(expected_event, builder.build())


if __name__ == '__main__':
  absltest.main()



# Copyright 2022, The TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for mechanism_calibration."""

from absl.testing import absltest
from absl.testing import parameterized
import attr
import numpy as np

from dp_accounting import dp_event
from dp_accounting import mechanism_calibration
from dp_accounting import privacy_accountant


@attr.define
class FakeEvent(dp_event.DpEvent):
  param: float


class FakeAccountant(privacy_accountant.PrivacyAccountant):

  def __init__(self, value_to_epsilon):
    super().__init__(
        privacy_accountant.NeighboringRelation.ADD_OR_REMOVE_ONE)
    self._value = 0.0
    self._value_to_epsilon = value_to_epsilon

  def _maybe_compose(self, event: dp_event.DpEvent, count: int,
                     do_compose: bool):
    self._value = event.param

  def get_epsilon(self, target_delta: float) -> float:
    return self._value_to_epsilon(self._value)


class MechanismCalibrationTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('identity', lambda x: x, 2.0),
      ('4_minus_x', lambda x: 4 - x, 2.0),
      ('square', np.square, np.sqrt(2)),
      ('cbrt', np.cbrt, 8.0),
      ('cubic', lambda x: (x - 5) ** 3 + 2, 5),
      ('trig_one', lambda x: np.cos(x / 3) + 2, 3 * np.pi / 2),
      ('trig_two', lambda x: np.sin(x - 5) + (x + 3) / 4, 5),
      ('trig_three', lambda x: (13 - x) / 4 - np.sin(x - 5), 5),
  )
  def test_basic_inversion(self, eps_fn, expected):

    value = mechanism_calibration.calibrate_dp_mechanism(
        make_fresh_accountant=lambda: FakeAccountant(eps_fn),
        make_event_from_param=FakeEvent,
        target_epsilon=2,
        target_delta=0,
        bracket_interval=mechanism_calibration.ExplicitBracketInterval(0, 10),
        tol=1e-12,
    )

    self.assertIsInstance(value, float)
    self.assertAlmostEqual(value, expected)

    accountant = FakeAccountant(eps_fn)
    accountant.compose(FakeEvent(value))
    epsilon = accountant.get_epsilon(0)
    self.assertLessEqual(epsilon, 2)

  @parameterized.named_parameters(
      ('neg_one_pos_one', lambda x: -1 if x < 0 else 1),
      ('pos_one_neg_one', lambda x: 1 if x < 0 else -1),
      ('sawtooth', lambda x: x - 1 if x < 0 else x + 1),
      ('neg_sawtooth', lambda x: -2 - x if x < 0 else 2 - x),
      ('x_plus_two', lambda x: x + 2 if x < 0 else x - 2),
      ('neg_one_minus_x', lambda x: 1 - x if x < 0 else -1 - x),
  )
  def test_discontinuous(self, eps_fn):
    value = mechanism_calibration.calibrate_dp_mechanism(
        make_fresh_accountant=lambda: FakeAccountant(eps_fn),
        make_event_from_param=FakeEvent,
        target_epsilon=0,
        target_delta=0,
        bracket_interval=mechanism_calibration.ExplicitBracketInterval(-1, 1),
        tol=1e-12,
    )

    self.assertIsInstance(value, float)
    self.assertAlmostEqual(value, 0)

    accountant = FakeAccountant(eps_fn)
    accountant.compose(FakeEvent(value))
    epsilon = accountant.get_epsilon(0)
    self.assertLessEqual(epsilon, 0)

  @parameterized.named_parameters(
      ('x_minus_2', lambda x: x - 2, 0),
      ('x_minus_2_1', lambda x: x - 2.1, -0.1),
      ('x_minus_2_9', lambda x: x - 2.9, -0.9),
      ('2_minus_x', lambda x: 2 - x, 0),
      ('1_9_minus_x', lambda x: 1.9 - x, -0.1),
      ('1_1_minus_x', lambda x: 1.1 - x, -0.9),
  )
  def test_discrete(self, eps_fn, expected_eps):
    value = mechanism_calibration.calibrate_dp_mechanism(
        make_fresh_accountant=lambda: FakeAccountant(eps_fn),
        make_event_from_param=FakeEvent,
        target_epsilon=0,
        target_delta=0,
        bracket_interval=mechanism_calibration.ExplicitBracketInterval(0, 5),
        discrete=True,
    )

    self.assertIsInstance(value, int)
    self.assertEqual(value, 2)

    accountant = FakeAccountant(eps_fn)
    accountant.compose(FakeEvent(value))
    epsilon = accountant.get_epsilon(0)
    self.assertAlmostEqual(epsilon, expected_eps)

  @parameterized.named_parameters(
      ('identity', lambda x: x, -1, -0.5),
      ('neg_x', lambda x: -x, -1, -0.5),
      ('exp_minus_2', lambda x: np.exp(x) - 2, 0, 0.1),
      ('1_minus_sqrt_x', lambda x: 1 - np.sqrt(x), 0, 0.1),
      ('log_minus_20', lambda x: np.log(x) - 20, 1, 2),
  )
  def test_search_for_explicit_bracket_interval(
      self, epsilon_gap, lower, guess
  ):
    interval = mechanism_calibration._search_for_explicit_bracket_interval(
        mechanism_calibration.LowerEndpointAndGuess(lower, guess),
        epsilon_gap,
    )
    lower_value = epsilon_gap(interval.endpoint_1)
    upper_value = epsilon_gap(interval.endpoint_2)
    self.assertLessEqual(lower_value * upper_value, 0)

  def test_raises_unknown_bracket_interval_type(self):
    class UnknownBracketInterval(mechanism_calibration.BracketInterval):
      pass

    with self.assertRaisesRegex(TypeError, 'Unrecognized bracket_interval'):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant=lambda: FakeAccountant(lambda x: x),
          make_event_from_param=FakeEvent,
          target_epsilon=1.0,
          target_delta=0,
          bracket_interval=UnknownBracketInterval(),
      )

  @parameterized.named_parameters(
      ('square', np.square, np.sqrt),
      ('cbrt', np.cbrt, lambda x: x**3),
      ('80th_power', lambda x: x**80, lambda x: x ** (1 / 80)),
      ('80th_root', lambda x: x ** (1 / 80), lambda x: x**80),
      (
          'cos',
          lambda x: np.cos(np.pi / 2 * x),
          lambda x: 2 / np.pi * np.arccos(x),
      ),
  )
  def test_bisect_finds_root_with_negative_value(self, function, inv_function):
    # Repeat the test many times to ensure negative value always returned.
    repetitions = 20
    generator = np.random.default_rng(seed=0xBAD5EED)
    for target in generator.uniform(size=repetitions):
      # pylint: disable=cell-var-from-loop
      root = mechanism_calibration._bisect(
          lambda x: function(x) - target, 0, 1, 1e-12
      )
      # pylint: enable=cell-var-from-loop
      expected = inv_function(target)
      self.assertAlmostEqual(root, expected)
      self.assertLessEqual(function(root), target)

  @parameterized.named_parameters(
      ('both_positive', lambda x: 1),
      ('both_negative', lambda x: -1),
  )
  def test_bisect_raises_values_same_sign(self, function):
    with self.assertRaisesRegex(ValueError, 'opposite signs'):
      mechanism_calibration._bisect(function, 0, 1, 1e-12)

  def test_raises_mfa_not_callable(self):
    with self.assertRaisesRegex(
        TypeError, 'make_fresh_accountant must be callable'
    ):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant='not a callable',
          make_event_from_param=FakeEvent,
          target_epsilon=1.0,
          target_delta=0,
          bracket_interval=mechanism_calibration.ExplicitBracketInterval(0, 5),
      )

  def test_raises_mefp_not_callable(self):
    with self.assertRaisesRegex(
        TypeError, 'make_event_from_param must be callable'
    ):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant=lambda: FakeAccountant(lambda x: x),
          make_event_from_param='not a callable',
          target_epsilon=1.0,
          target_delta=0,
          bracket_interval=mechanism_calibration.ExplicitBracketInterval(0, 5),
      )

  def test_raises_target_epsilon_negative(self):
    with self.assertRaisesRegex(ValueError, 'nonnegative'):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant=lambda: FakeAccountant(lambda x: x),
          make_event_from_param=FakeEvent,
          target_epsilon=-1.0,
          target_delta=0,
          bracket_interval=mechanism_calibration.ExplicitBracketInterval(0, 5),
      )

  def test_raises_target_delta_out_of_range(self):
    with self.assertRaisesRegex(ValueError, 'in range'):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant=lambda: FakeAccountant(lambda x: x),
          make_event_from_param=FakeEvent,
          target_epsilon=0.0,
          target_delta=-0.1,
          bracket_interval=mechanism_calibration.ExplicitBracketInterval(0, 5),
      )

    with self.assertRaisesRegex(ValueError, 'in range'):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant=lambda: FakeAccountant(lambda x: x),
          make_event_from_param=FakeEvent,
          target_epsilon=0.0,
          target_delta=1.1,
          bracket_interval=mechanism_calibration.ExplicitBracketInterval(0, 5),
      )

  def test_bad_bracket_interval(self):
    with self.assertRaisesRegex(ValueError, 'did not bracket a solution'):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant=lambda: FakeAccountant(lambda x: x),
          make_event_from_param=FakeEvent,
          target_epsilon=1.0,
          target_delta=0.0,
          bracket_interval=mechanism_calibration.ExplicitBracketInterval(2, 5),
      )

    with self.assertRaisesRegex(ValueError, 'did not bracket a solution'):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant=lambda: FakeAccountant(lambda x: x),
          make_event_from_param=FakeEvent,
          target_epsilon=1.0,
          target_delta=0.0,
          bracket_interval=mechanism_calibration.ExplicitBracketInterval(-2, 0),
      )

    with self.assertRaisesRegex(ValueError, 'must be less than'):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant=lambda: FakeAccountant(lambda x: x),
          make_event_from_param=FakeEvent,
          target_epsilon=1.0,
          target_delta=0.0,
          bracket_interval=mechanism_calibration.LowerEndpointAndGuess(2, 0),
      )

  def test_negative_tol(self):
    with self.assertRaisesRegex(ValueError, 'tol'):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant=lambda: FakeAccountant(lambda x: x),
          make_event_from_param=FakeEvent,
          target_epsilon=1.0,
          target_delta=0.0,
          bracket_interval=mechanism_calibration.LowerEndpointAndGuess(0, 1),
          tol=-1,
      )

  def test_no_bracket_interval_found(self):
    with self.assertRaises(mechanism_calibration.NoBracketIntervalFoundError):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant=lambda: FakeAccountant(lambda x: x),
          make_event_from_param=FakeEvent,
          target_epsilon=1.0e10,
          target_delta=0.0,
          bracket_interval=mechanism_calibration.LowerEndpointAndGuess(0, 1),
      )

  def test_nonempty_accountant(self):
    def make_fresh_accountant():
      accountant = FakeAccountant(lambda x: x)
      accountant.compose(FakeEvent(1.0))
      return accountant

    with self.assertRaises(mechanism_calibration.NonEmptyAccountantError):
      mechanism_calibration.calibrate_dp_mechanism(
          make_fresh_accountant=make_fresh_accountant,
          make_event_from_param=FakeEvent,
          target_epsilon=0.5,
          target_delta=0.0,
          bracket_interval=mechanism_calibration.ExplicitBracketInterval(0, 1),
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Basic Example for Using Privacy Loss Distributions."""

from absl import app

from dp_accounting.pld import privacy_loss_distribution


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  # The parameter of Laplace Noise added
  parameter_laplace = 3
  # PLD for one execution of the Laplace Mechanism. (Throughout we assume that
  # sensitivity = 1.)
  laplace_pld = privacy_loss_distribution.from_laplace_mechanism(
      parameter_laplace, value_discretization_interval=1e-3
  )

  # Number of times Laplace Mechanism is run
  num_laplace = 40
  # PLD for num_laplace executions of the Laplace Mechanism.
  composed_laplace_pld = laplace_pld.self_compose(num_laplace)

  epsilon = 10
  delta = composed_laplace_pld.get_delta_for_epsilon(epsilon)
  print(
      'An algorithm that executes the Laplace Mechanism with parameter '
      f'{parameter_laplace} for a total of {num_laplace} times is '
      f'({epsilon}, {delta})-DP.'
  )

  # PLDs for different mechanisms can also be composed. Below is an example in
  # which we compose PLDs for Laplace Mechanism and Gaussian Mechanism.

  # STD of the Gaussian Noise
  standard_deviation = 5
  # PLD for an execution of the Gaussian Mechanism.
  gaussian_pld = privacy_loss_distribution.from_gaussian_mechanism(
      standard_deviation, value_discretization_interval=1e-3
  )

  # PLD for num_laplace executions of the Laplace Mechanism and one execution of
  # the Gaussian Mechanism.
  composed_laplace_and_gaussian_pld = composed_laplace_pld.compose(gaussian_pld)

  epsilon = 10
  delta = composed_laplace_and_gaussian_pld.get_delta_for_epsilon(epsilon)
  print(
      'An algorithm that executes the Laplace Mechanism with parameter '
      f'{parameter_laplace} for a total of {num_laplace} times and in '
      'addition executes once the Gaussian Mechanism with STD '
      f'{standard_deviation} is ({epsilon}, {delta})-DP.'
  )


if __name__ == '__main__':
  app.run(main)



# Copyright 2020 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup for DP Accounting package."""

import os
import setuptools

here = os.path.dirname(os.path.abspath(__file__))


def _get_version():
  """Parses the version number from VERSION file."""
  with open(os.path.join(here, "VERSION")) as f:
    try:
      version_line = next(
          line for line in f if not line.startswith("\"\"\""))
    except StopIteration:
      raise ValueError("Version not defined in VERSION")
    else:
      return version_line.strip('\n \'"')


def _parse_requirements(path):
  """Parses requirements from file."""
  with open(os.path.join(here, path)) as f:
    deps = []
    for line in f:
      if line.startswith("dataclasses"):
        # For python version 3.7 onwards, dataclasses module is already included
        # as part of the core library.
        deps.append("dataclasses; python_version < \'3.7\'")
      elif not (line.isspace() or line.startswith("#")):
        deps.append(line.rstrip())
    return deps


def _read_description(path):
  """Read the description from README file."""
  # Exclude example from package description.
  return open(os.path.join(here, path)).read().split("## Examples")[0]

setuptools.setup(
    name="dp-accounting",
    version=_get_version(),
    author="Google Differential Privacy Team",
    author_email="dp-open-source@google.com",
    description="Tools for tracking differential privacy budgets",
    long_description=_read_description("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/google/differential-privacy/",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=_parse_requirements("requirements.txt"),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.7",
    license="Apache 2.0",
    keywords="differential-privacy accounting",
)



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Abstract base class for tests of `PrivacyAccountant` classes.

Checks that a class derived from `PrivacyAccountant` has the correct behavior
for standard `DpEvent` classes.
"""

from typing import Collection

from absl.testing import absltest
from absl.testing import parameterized

from dp_accounting import dp_event
from dp_accounting import privacy_accountant


class UnknownDpEvent(dp_event.DpEvent):
  pass


@absltest.skipThisClass('only intended to be run by subclasses')
class PrivacyAccountantTest(parameterized.TestCase):

  def _make_test_accountants(
      self) -> Collection[privacy_accountant.PrivacyAccountant]:
    """Makes a list of accountants to test.

    Subclasses should define this to return a list of accountants to be tested.

    Returns:
      A list of accountants to test.
    """
    return []

  def test_make_test_accountants(self):
    self.assertNotEmpty(self._make_test_accountants())

  @parameterized.product(
      unsupported_event=(dp_event.UnsupportedDpEvent(), UnknownDpEvent()),
      nest_fn=(
          lambda event_: event_,
          lambda event_: dp_event.ComposedDpEvent([event_]),
          lambda event_: dp_event.SelfComposedDpEvent(event_, 10),
      ),
  )
  def test_unsupported(self, unsupported_event, nest_fn):
    event_ = nest_fn(unsupported_event)
    for accountant in self._make_test_accountants():
      composition_error = accountant._maybe_compose(
          event_, count=1, do_compose=False
      )
      self.assertIsNotNone(composition_error)
      self.assertEqual(composition_error.invalid_event, unsupported_event)
      self.assertFalse(accountant.supports(event_))
      with self.assertRaisesRegex(
          privacy_accountant.UnsupportedEventError, 'caused by subevent'
      ):
        accountant.compose(event_)

  def test_no_events(self):
    for accountant in self._make_test_accountants():
      self.assertEqual(accountant.get_epsilon(1e-12), 0)
      self.assertEqual(accountant.get_epsilon(0), 0)
      self.assertEqual(accountant.get_epsilon(1), 0)
      try:
        self.assertEqual(accountant.get_delta(1e-12), 0)
        self.assertEqual(accountant.get_delta(0), 0)
        self.assertEqual(accountant.get_delta(float('inf')), 0)
      except NotImplementedError:
        # Implementing `get_delta` is optional.
        pass

  def test_no_op(self):
    for accountant in self._make_test_accountants():
      event = dp_event.NoOpDpEvent()
      self.assertTrue(accountant.supports(event))
      accountant.compose(event)
      self.assertEqual(accountant.get_epsilon(1e-12), 0)
      self.assertEqual(accountant.get_epsilon(0), 0)
      self.assertEqual(accountant.get_epsilon(1), 0)
      try:
        self.assertEqual(accountant.get_delta(1e-12), 0)
        self.assertEqual(accountant.get_delta(0), 0)
        self.assertEqual(accountant.get_delta(float('inf')), 0)
      except NotImplementedError:
        # Implementing `get_delta` is optional.
        pass

  @parameterized.parameters(
      dp_event.NonPrivateDpEvent(),
      dp_event.ComposedDpEvent([dp_event.NonPrivateDpEvent()]),
      dp_event.ComposedDpEvent(
          [dp_event.NoOpDpEvent(), dp_event.NonPrivateDpEvent()]
      ),
      dp_event.SelfComposedDpEvent(dp_event.NonPrivateDpEvent(), 10),
  )
  def test_non_private(self, non_private_event):
    for accountant in self._make_test_accountants():
      self.assertTrue(accountant.supports(non_private_event))
      accountant.compose(non_private_event)
      self.assertEqual(accountant.get_epsilon(0.99), float('inf'))
      self.assertEqual(accountant.get_epsilon(0), float('inf'))
      self.assertEqual(accountant.get_epsilon(1), float('inf'))
      try:
        self.assertEqual(accountant.get_delta(100), 1)
        self.assertEqual(accountant.get_delta(0), 1)
        self.assertEqual(accountant.get_delta(float('inf')), 1)
      except NotImplementedError:
        # Implementing `get_delta` is optional.
        pass



# Copyright 2022, The TensorFlow Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library for calibration of differentially private mechanisms.

Algorithms to optimize some quantity while remaining within a specified privacy
budget.
"""

from typing import Callable, Optional, Union

import attr
from scipy import optimize

from dp_accounting import dp_event
from dp_accounting import privacy_accountant


class BracketInterval(object):
  pass


@attr.define(frozen=True)
class ExplicitBracketInterval(BracketInterval):
  endpoint_1: float
  endpoint_2: float


@attr.define(frozen=True)
class LowerEndpointAndGuess(BracketInterval):
  lower_endpoint: float
  initial_guess: float


class NoBracketIntervalFoundError(Exception):
  """Error raised when explicit bracket interval cannot be found."""


class NonEmptyAccountantError(Exception):
  """Error raised when result of make_fresh_accountant has nonempty ledger."""


def _search_for_explicit_bracket_interval(
    bracket_interval: LowerEndpointAndGuess,
    epsilon_gap: Callable[[float], float]) -> ExplicitBracketInterval:
  """Explores exponentially increasing interval to find an explicit bracket.

  Args:
    bracket_interval: A LowerEndpointAndGuess which will be expanded to find
      an explicit interval.
    epsilon_gap: Function computing the epsilon at the provided value minus
      the target epsilon. It is assumed that this function is monotonic with
      respect to its parameter, otherwise the search could fail.

  Returns:
    A valid ExplicitBracketInterval.

  Raises:
    NoBracketIntervalFoundError: if no valid bracketing interval is found
      within a factor of 2**30 of the initial guess.
  """
  lower, upper = attr.astuple(bracket_interval)
  if lower >= upper:
    raise ValueError(
        f'bracket_interval.lower_endpoint ({bracket_interval.lower_endpoint}) '
        f'must be less than bracket_interval.initial_guess '
        f'({bracket_interval.initial_guess}).')

  lower_value = epsilon_gap(lower)
  upper_value = epsilon_gap(upper)

  gap = upper - lower
  num_tries = 0

  while lower_value * upper_value > 0:
    num_tries += 1
    if num_tries > 30:
      raise NoBracketIntervalFoundError(
          'Unable to find bracketing interval within 2**30 of initial guess. '
          'Consider providing an ExplicitBracketInterval.')

    gap *= 2  # Loop invariant: gap = initial_gap * (2 ** num_tries).
    lower, upper = upper, upper + gap
    lower_value, upper_value = upper_value, epsilon_gap(upper)

  return ExplicitBracketInterval(lower, upper)


def _bisect(
    function: Callable[[float], float],
    lower: float,
    upper: float,
    tol: float,
    lower_value: Optional[float] = None,
    upper_value: Optional[float] = None,
) -> float:
  """Bisection search to find approximate root with non-positive value.

  Args:
    function: Function to find approximate root of. Conceptually, the function
      should be continuous, although really the only requirement is that it has
      opposite signs at the endpoints.
    lower: Lower endpoint.
    upper: Upper endpoint.
    tol: Terminate when endpoints are within tol of each other.
    lower_value: Value at lower endpoint.
    upper_value: Value at upper endpoint.

  Returns:
    A point with non-positive value within tol of root.

  Raises:
    ValueError: If values at lower and upper do not have opposite signs.
  """
  if lower_value is None:
    lower_value = function(lower)
  if upper_value is None:
    upper_value = function(upper)

  if lower_value == 0:
    return lower
  if upper_value == 0:
    return upper

  if lower_value * upper_value > 0:
    raise ValueError('Values must have opposite signs.')

  if upper - lower <= tol:
    return lower if lower_value < 0 else upper

  middle = (lower + upper) / 2
  middle_value = function(middle)

  if middle_value == 0:
    return middle
  elif lower_value * middle_value < 0:
    return _bisect(function, lower, middle, tol, lower_value, middle_value)
  else:
    return _bisect(function, middle, upper, tol, middle_value, upper_value)


def calibrate_dp_mechanism(
    make_fresh_accountant: Callable[[], privacy_accountant.PrivacyAccountant],
    make_event_from_param: Union[Callable[[float], dp_event.DpEvent],
                                 Callable[[int], dp_event.DpEvent]],
    target_epsilon: float,
    target_delta: float,
    bracket_interval: Optional[BracketInterval] = None,
    discrete: bool = False,
    tol: Optional[float] = None) -> Union[float, int]:
  """Searches for optimal mechanism parameter value within privacy budget.

  The procedure searches over the space of parameters by creating, for each
  sample value, a DpEvent representing the mechanism generated from that value,
  and a freshly initialized PrivacyAccountant. Then the accountant is applied to
  the event to determine its epsilon at the target delta. Brent's method is used
  to determine the value of the parameter at which the target epsilon is
  achieved.

  Args:
    make_fresh_accountant: A callable with no parameters that returns an
      initialized PrivacyAccountant. The accountants that are returned across
      multiple calls are assumed to be initialized identically. It is an error
      for the initialized accountant's `ledger` property to return anything
      besides `NoOpDpEvent`.
    make_event_from_param: A callable that takes a parameter value as an
      argument and creates a `DpEvent` representing the mechanism defined using
      that value.
    target_epsilon: The target epsilon value.
    target_delta: The target delta value.
    bracket_interval: A BracketInterval used to determine the upper and lower
      endpoints of the interval within which Brent's method will search. If
      None, searches for a non-negative bracket starting from [0, 1].
    discrete: A bool determining whether the parameter is continuous or discrete
      valued. If True, the parameter is assumed to take only integer values.
      Concretely, `discrete=True` has three effects. 1) ints, not floats are
      passed to `make_event_from_param`. 2) The minimum optimization tolerance
      is 0.5. 3) An integer is returned.
    tol: The tolerance, in parameter space. If the maximum (or minimum) value of
      the parameter that meets the privacy requirements is x*,
      calibrate_dp_mechanism is guaranteed to return a value x such that |x -
      x*| <= tol. If `None`, tol is set to 1e-6 for continuous parameters or 0.5
      for discrete parameters.

  Returns:
    A value of the parameter within tol of the optimum subject to the privacy
    constraint. If discrete=True, the returned value will be an integer.
    Otherwise it will be a float.

  Raises:
    NoBracketIntervalFoundError: if bracket_interval is LowerEndpointAndGuess
      and no upper bound can be found within a factor of 2**30 of the original
      guess.
    NonEmptyAccountantError: if make_fresh_accountant returns an accountant with
      nonempty ledger.
  """

  if not callable(make_fresh_accountant):
    raise TypeError(f'make_fresh_accountant must be callable. '
                    f'found {type(make_fresh_accountant)}.')

  if not callable(make_event_from_param):
    raise TypeError(f'make_event_from_param must be callable. '
                    f'found {type(make_event_from_param)}.')

  if target_epsilon < 0:
    raise ValueError(f'target_epsilon must be nonnegative. Found '
                     f'{target_epsilon}.')

  if not 0 <= target_delta <= 1:
    raise ValueError(f'target_delta must be in range [0, 1]. Found '
                     f'{target_delta}.')

  if bracket_interval is None:
    bracket_interval = LowerEndpointAndGuess(0, 1)

  if tol is None:
    tol = 1.0 if discrete else 1e-6
  elif discrete:
    tol = max(tol, 1.0)
  elif tol <= 0:
    raise ValueError(f'tol must be positive. Found {tol}.')

  def epsilon_gap(x: float) -> float:
    if discrete:
      x = round(x)
    event = make_event_from_param(x)
    accountant = make_fresh_accountant()
    if not isinstance(accountant.ledger, dp_event.NoOpDpEvent):
      raise NonEmptyAccountantError()
    return accountant.compose(event).get_epsilon(target_delta) - target_epsilon

  if isinstance(bracket_interval, LowerEndpointAndGuess):
    bracket_interval = _search_for_explicit_bracket_interval(
        bracket_interval, epsilon_gap)
  elif not isinstance(bracket_interval, ExplicitBracketInterval):
    raise TypeError(f'Unrecognized bracket_interval type: '
                    f'{type(bracket_interval)}')

  try:
    root, result = optimize.brentq(
        epsilon_gap,
        bracket_interval.endpoint_1,
        bracket_interval.endpoint_2,
        xtol=tol,
        full_output=True)
  except ValueError as err:
    raise ValueError(
        '`brentq` raised ValueError. This often means the supplied bracket '
        f'interval {bracket_interval} did not bracket a solution.') from err

  if not result.converged:
    root = None
  else:
    # We need to ensure that gap is not positive, guaranteeing the returned
    # parameter gives no less privacy than was requested. Since epsilon_gap can
    # be expensive to compute, we first try values near the returned root.
    # Considering brentq's contract that there exists a root within tol of the
    # returned value, in most cases adding +/- tol will suffice.
    if epsilon_gap(root) > 0:
      if epsilon_gap(root + tol) < 0:
        root += tol
      elif epsilon_gap(root - tol) < 0:
        root -= tol
      else:
        root = None

  if root is None:
    # Fallback to custom bisection that guarantees root with non-positive value.
    root = _bisect(
        epsilon_gap,
        bracket_interval.endpoint_1,
        bracket_interval.endpoint_2,
        tol,
    )

  if discrete:
    root = round(root)

  return root
